package highscoretetrisai.real;

import java.util.*;
import static java.lang.Math.*;
import static highscoretetrisai.real.Playfield.*;

public class Tetrimino {
  
  public static final Tetrimino[] T = {
    new Tetrimino(new int[][] { { -1,  0 }, {  0,  0 }, {  1,  0 }, {  0,  1 }, }),  // down (spawn)
    new Tetrimino(new int[][] { {  0, -1 }, { -1,  0 }, {  0,  0 }, {  0,  1 }, }),  // left
    new Tetrimino(new int[][] { { -1,  0 }, {  0,  0 }, {  1,  0 }, {  0, -1 }, }),  // up
    new Tetrimino(new int[][] { {  0, -1 }, {  0,  0 }, {  1,  0 }, {  0,  1 }, }),  // right
  };
  
  public static final Tetrimino[] J = {
    new Tetrimino(new int[][] { { -1,  0 }, {  0,  0 }, {  1,  0 }, {  1,  1 }, }),  // down (spawn)
    new Tetrimino(new int[][] { {  0, -1 }, {  0,  0 }, { -1,  1 }, {  0,  1 }, }),  // left
    new Tetrimino(new int[][] { { -1, -1 }, { -1,  0 }, {  0,  0 }, {  1,  0 }, }),  // up
    new Tetrimino(new int[][] { {  0, -1 }, {  1, -1 }, {  0,  0 }, {  0,  1 }, }),  // right  
  };
  
  public static final Tetrimino[] Z = {
    new Tetrimino(new int[][] { { -1,  0 }, {  0,  0 }, {  0,  1 }, {  1,  1 }, }),  // horizontal (spawn) 
    new Tetrimino(new int[][] { {  1, -1 }, {  0,  0 }, {  1,  0 }, {  0,  1 }, }),  // vertical
  };
  
  public static final Tetrimino[] O = {
    new Tetrimino(new int[][] { { -1,  0 }, {  0,  0 }, { -1,  1 }, {  0,  1 }, }),  // spawn
  };
  
  public static final Tetrimino[] S = {
    new Tetrimino(new int[][] { {  0,  0 }, {  1,  0 }, { -1,  1 }, {  0,  1 }, }),  // horizontal (spawn)
    new Tetrimino(new int[][] { {  0, -1 }, {  0,  0 }, {  1,  0 }, {  1,  1 }, }),  // vertical
  };
     
  public static final Tetrimino[] L = {
    new Tetrimino(new int[][] { { -1,  0 }, {  0,  0 }, {  1,  0 }, { -1,  1 }, }),  // down (spawn)
    new Tetrimino(new int[][] { { -1, -1 }, {  0, -1 }, {  0,  0 }, {  0,  1 }, }),  // left
    new Tetrimino(new int[][] { {  1, -1 }, { -1,  0 }, {  0,  0 }, {  1,  0 }, }),  // up
    new Tetrimino(new int[][] { {  0, -1 }, {  0,  0 }, {  0,  1 }, {  1,  1 }, }),  // right
  };

  public static final Tetrimino[] I = {
    new Tetrimino(new int[][] { { -2,  0 }, { -1,  0 }, {  0,  0 }, {  1,  0 }, }),  // horizontal (spawn)
    new Tetrimino(new int[][] { {  0, -2 }, {  0, -1 }, {  0,  0 }, {  0,  1 }, }),  // vertical       
  };  
  
  public static final Tetrimino[][] TETRIMINOS = { T, J, Z, O, S, L, I, };
  
  public final Offset[] offsets = new Offset[4];
  public final Offset[] supports;
  public final boolean[][] validPosition = new boolean[HEIGHT][WIDTH];
  public final int minOffsetX;
  public final int maxOffsetX;
  public final int minOffsetY;
  public final int maxOffsetY;
  
  public Tetrimino(final int[][] squares) {
    int minX = Integer.MAX_VALUE;
    int maxX = Integer.MIN_VALUE;
    int minY = Integer.MAX_VALUE;
    int maxY = Integer.MIN_VALUE;
    for(int i = squares.length - 1; i >= 0; --i) {
      final Offset offset = this.offsets[i] = new Offset(squares[i]);
      minX = min(minX, offset.x);
      maxX = max(maxX, offset.x);
      minY = min(minY, offset.y);
      maxY = max(maxY, offset.y);
    }
    this.minOffsetX = minX;    
    this.maxOffsetX = maxX;
    this.minOffsetY = minY;
    this.maxOffsetY = maxY;
    for(int y = HEIGHT - 1; y >= 0; --y) {
      for(int x = WIDTH - 1; x >= 0; --x) {
        validPosition[y][x] = true;
        for(final Offset offset : this.offsets) {
          final int X = offset.x + x;
          final int Y = offset.y + y;
          if (X < 0 || X >= WIDTH || Y >= HEIGHT) {
            validPosition[y][x] = false;
            break;
          }
        }
      }
    }
    
    final boolean[][] matrix = new boolean[5][5];
    for(final Offset offset : this.offsets) {
      matrix[offset.y + 2][offset.x + 2] = true;
    }
    final List<Offset> supportList = new ArrayList<>();
    for(int y = 4; y > 0; --y) {
      for(int x = 4; x >= 0; --x) {
        if (!matrix[y][x] && matrix[y - 1][x]) {
          supportList.add(new Offset(x - 2, y - 2));
        }
      }
    }
    supports = new Offset[supportList.size()];
    supportList.toArray(supports);
  }
  
  public int getLockHeight(final int y) {
    return HEIGHT - (y + maxOffsetY);
  }
}